@@ -20,6 +20,7 @@
 #include <unistd.h>
 #include <getopt.h>
 #include <ctype.h>
+#include <regex.h>
 #include <libical/ical.h>
 /*
 #include <icaltime.h>
@@ -35,6 +36,7 @@
 #define ERR_USAGE 2
 #define ERR_DATE 3
 #define ERR_PARSE 4
+#define ERR_BUG 5
 
 #define USAGE "Usage: ical2html [options] start duration [file]\n\
   -p, --class=CLASS            only (PUBLIC, CONFIDENTIAL, PRIVATE, NONE)\n\
@@ -43,8 +45,11 @@
   -C, --not-category=CATEGORY  exclude events of this category\n\
   -d, --description            include event's long description in a <PRE>\n\
   -l, --location               include event's location in that <PRE>\n\
+  -t, --title=TEXT             use text for HTML title\n\
   -f, --footer=TEXT            add text at the bottom of the HTML file\n\
+  -T, --today                  mark current day with #today in HTML file\n\
   -z, --timezone=country/city  adjust for this timezone (default: GMT)\n\
+  -m, --monday                 draw Monday as first week day (Sunday is default)\n\
   start is of the form yyyymmdd, e.g., 20020927 (27 Sep 2002)\n\
   duration is in days or weeks, e.g., P5W (5 weeks) or P60D (60 days)\n\
   file is an iCalendar file, default is standard input\n"
@@ -57,12 +62,15 @@ static struct option options[] = {
   {"not-category", 1, 0, 'C'},
   {"description", 0, 0, 'd'},
   {"location", 0, 0, 'l'},
+  {"title", 1, 0, 't'},
   {"footer", 1, 0, 'f'},
+  {"today", 0, 0, 'T'},
   {"timezone", 1, 0, 'z'},
+  {"monday", 0, 0, 'm'},
   {0, 0, 0, 0}
 };
 
-#define OPTIONS "dlp:P:c:C:f:z:"
+#define OPTIONS "dlmTp:P:c:C:f:z:"
 
 static const char *months[] = {"", "January", "February", "March", "April",
 			       "May", "June", "July", "August", "September",
@@ -72,6 +80,8 @@ static const char *months[] = {"", "January", "February", "March", "April",
 typedef struct _event_item {
   struct icaltimetype start;
   struct icaltimetype end;
+  const char *uid;
+  struct icaltimetype recur_time;
   icalcomponent *event;
 } event_item;
 
@@ -99,6 +109,8 @@ static int compare_events(const void *aa, const void *bb)
 
   h = icaltime_compare(a->start, b->start);
   if (h == 0) h = icaltime_compare(a->end, b->end);
+  if (h == 0) h = strcmp(a->uid, b->uid);
+  if (h == 0) h = icaltime_compare(a->recur_time, b->recur_time);
   return h;
 }
 
@@ -110,19 +122,85 @@ static char* read_stream(char *s, size_t size, void *d)
 } 
 
 
+/* print_esc -- print with <, >, & and " escaped, until end or \0 */
+static void print_esc(const char *s, const char *end)
+{
+  const char *t;
+
+  for (t = s; *t && t != end; t++)
+    switch (*t) {
+    case '<': printf("&lt;"); break;
+    case '>': printf("&gt;"); break;
+    case '&': printf("&amp;"); break;
+    default: putchar(*t);
+    }
+}
+
+
+/* print_escaped -- print with <, >, & and " escaped */
+static void print_escaped(const char *s)
+{
+  print_esc(s, NULL);
+}
+
+
+/* print_escaped_with_links -- print with <, >, & and " escaped, URLs linked */
+static void print_escaped_with_links(const char *s)
+{
+  static regex_t url_pattern;
+  static int initialized = 0;
+  regmatch_t pmatch[1];
+  const char *t;
+
+  if (!initialized) {
+    if (regcomp(&url_pattern,
+	"^([a-z]+://|mailto:|geo:|urn:[a-z0-9-]+:)([][a-z0-9._~/?#[@!$&'()*+,;=-]|%[0-9a-f][0-9a-f])+",
+	REG_EXTENDED | REG_ICASE) != 0)
+      fatal(ERR_BUG, "regcomp() failed. This is a bug in the program. Sorry!\n");
+    initialized = 1;
+  }
+
+  for (t = s; *t; t++)
+    /* TODO: handle case that regexec returns REG_ESPACE or REG_ASSERT. */
+    if (regexec(&url_pattern, t, 1, pmatch, 0) != REG_NOMATCH) {
+      printf("<a href=\"");
+      print_esc(t, t + pmatch[0].rm_eo);
+      printf("\">");
+      print_esc(t, t + pmatch[0].rm_eo);
+      printf("</a>");
+      t += pmatch[0].rm_eo - 1;
+    } else {
+      switch (*t) {
+      case '<': printf("&lt;"); break;
+      case '>': printf("&gt;"); break;
+      case '&': printf("&amp;"); break;
+      default: putchar(*t);
+      }
+    }
+}
+
+
 /* print_header -- print boilerplate at start of output */
-static void print_header(struct icaltimetype start, struct icaldurationtype dur)
+static void print_header(struct icaltimetype start, struct icaldurationtype dur,
+			 const char *title)
 {
   struct icaltimetype end = icaltime_add(start, dur);
 
-  printf("<!doctype html public \"-//W3C//DTD HTML 4.01//EN\"\n");
-  printf("  \"http://www.w3.org/TR/html4/strict.dtd\">\n");
-  printf("<meta http-equiv=\"Content-Type\" ");
-  printf("content=\"text/html;charset=UTF-8\">\n");
-  printf("<title>Calendar %d %s %d - %d %s %d</title>\n",
-	 start.day, months[start.month], start.year,
-	 end.day, months[end.month], end.year);
-  printf("<link rel=stylesheet href=\"calendar.css\">\n\n");
+  printf("<!DOCTYPE HTML>\n");
+  printf("<html>\n");
+  printf("<head>\n");
+  printf("<meta charset=\"UTF-8\" />\n");
+  printf("<title>");
+  if (title)
+    print_escaped(title);
+  else
+    printf("Calendar %d %s %d - %d %s %d",
+	   start.day, months[start.month], start.year,
+	   end.day, months[end.month], end.year);
+  printf("</title>\n");
+  printf("<link rel=\"stylesheet\" type=\"text/css\" href=\"calendar.css\" />\n");
+  printf("</head>\n");
+  printf("<body>\n");
 }
 
 
@@ -130,6 +208,8 @@ static void print_header(struct icaltimetype start, struct icaldurationtype dur)
 static void print_footer(const char *footer)
 {
   if (footer) printf("%s\n", footer);
+  printf("</body>\n");
+  printf("</html>\n");
 }
 
 
@@ -144,21 +224,6 @@ static void print_as_one_word(const char *s)
 }
 
 
-/* print_escaped -- printf with <, >, & and " escaped */
-static void print_escaped(const char *s)
-{
-  const char *t;
-
-  for (t = s; *t; t++) 
-    switch (*t) {
-    case '<': printf("&lt;"); break;
-    case '>': printf("&gt;"); break;
-    case '&': printf("&amp;"); break;
-    default: putchar(*t);
-    }
-}
-
-
 #if 0
 /* print_attribute -- printf with &, " and newlines escaped */
 static void print_attribute(const char *s)
@@ -184,9 +249,14 @@ static void print_event(const event_item ev, const int do_description, const int
   icaltimezone *utc = icaltimezone_get_utc_timezone();
   icaltimetype start_utc, end_utc;
   icalproperty *p, *desc, *loc;
+  enum icalproperty_status status;
   int first;
 
-  printf("<div class=vevent><p class=\"");
+  status = icalcomponent_get_status(ev.event);
+  printf("<div class=\"vevent");
+  if (status)
+    printf(" %s", icalproperty_status_to_string(status));
+  printf("\"><div class=\"");
 
   /* Add all categories to the class attribute */
   first = 1;
@@ -196,7 +266,7 @@ static void print_event(const event_item ev, const int do_description, const int
     print_as_one_word(icalproperty_get_categories(p));
     p = icalcomponent_get_next_property(ev.event, ICAL_CATEGORIES_PROPERTY);
   }
-  printf("\">\n<span class=categories>");
+  printf("\">\n<span class=\"categories\">");
 
   /* Also add all categories as content */
   first = 1;
@@ -212,8 +282,8 @@ static void print_event(const event_item ev, const int do_description, const int
   start_utc = icaltime_convert_to_zone(ev.start, utc);
   end_utc = icaltime_convert_to_zone(ev.end, utc);
   if (ev.start.hour || ev.start.minute || ev.end.hour || ev.end.minute)
-    printf("<span class=time><abbr class=dtstart\n\
-title=\"%04d%02d%02dT%02d%02d%02dZ\">%02d:%02d</abbr>-<abbr class=dtend\n\
+    printf("<span class=\"time\"><abbr class=\"dtstart\"\n\
+title=\"%04d%02d%02dT%02d%02d%02dZ\">%02d:%02d</abbr>-<abbr class=\"dtend\"\n\
 title=\"%04d%02d%02dT%02d%02d%02dZ\">%02d:%02d</abbr></span>\n",
 	   start_utc.year, start_utc.month, start_utc.day, start_utc.hour,
 	   start_utc.minute, start_utc.second,
@@ -222,15 +292,15 @@ title=\"%04d%02d%02dT%02d%02d%02dZ\">%02d:%02d</abbr></span>\n",
 	   end_utc.minute, end_utc.second,
 	   ev.end.hour, ev.end.minute);
   else
-    printf("<span class=notime><abbr class=dtstart\n\
-title=\"%04d%02d%02d\">(whole</abbr> <abbr class=duration\n\
+    printf("<span class=\"notime\"><abbr class=\"dtstart\"\n\
+title=\"%04d%02d%02d\">(whole</abbr> <abbr class=\"duration\"\n\
 title=\"1D\">day)</abbr></span>\n", start_utc.year, start_utc.month,
 	   start_utc.day);
 
   /* Print the summary */
-  printf("<span class=summary>");
+  printf("<span class=\"summary\">");
   p = icalcomponent_get_first_property(ev.event, ICAL_SUMMARY_PROPERTY);
-  if (p) print_escaped(icalproperty_get_summary(p));
+  if (p) print_escaped_with_links(icalproperty_get_summary(p));
   printf("</span>\n");
 
   /* If we want descriptions, check if there is one */
@@ -248,19 +318,19 @@ title=\"1D\">day)</abbr></span>\n", start_utc.year, start_utc.month,
   /* If we have a description and/or location, print them */
   if (desc || loc) printf("<pre>");
   if (desc) {
-    printf("<span class=description>");
-    print_escaped(icalproperty_get_description(desc));
+    printf("<span class=\"description\">");
+    print_escaped_with_links(icalproperty_get_description(desc));
     printf("</span>");
   }
   if (desc && loc) printf("\n");
   if (loc) {
-    printf("<b class=location>");
-    print_escaped(icalproperty_get_description(loc));
+    printf("<b class=\"location\">");
+    print_escaped_with_links(icalproperty_get_description(loc));
     printf("</b>");
   }
   if (desc || loc) printf("</pre>\n");
 
-  printf("</div>\n\n");
+  printf("</div></div>\n\n");
 }
 
 
@@ -268,13 +338,18 @@ title=\"1D\">day)</abbr></span>\n", start_utc.year, start_utc.month,
 static void print_calendar(const struct icaltimetype start,
 			   const struct icaldurationtype duration,
 			   const int nrevents, const event_item events[],
-			   const int do_description, const int do_location)
+			   const int do_description, const int do_location,
+			   const int starts_on_monday,
+			   const int do_today, const struct icaltimetype now)
 {
   struct icaltimetype day;
   struct icaltimetype end;
   char s[9];
   int y, m, d, w;
-  int i = 0;			/* Loop over events */
+  int i = 0;	/* Loop over events */
+  int skip;	/* How many days to skip of that week until 1st */
+  int lastDay;
+  if (starts_on_monday) lastDay = 2; else lastDay = 1;
 
   end = icaltime_add(start, duration);
 
@@ -289,15 +364,30 @@ static void print_calendar(const struct icaltimetype start,
       sprintf(s, "%04d%02d01", y, m);
       day = icaltime_from_string(s);
 
-      printf("<table><caption>%s %d</caption>\n", months[m], y);
-      printf("<thead><tr><th>Sunday<th>Monday<th>Tuesday<th>Wednesday");
-      printf("<th>Thursday<th>Friday<th>Saturday\n");
-      printf("<tbody>\n");
+      printf("<table border><caption>%s %d</caption>\n", months[m], y);
+      printf("<thead><tr>\n");
+      if (starts_on_monday) {
+         printf("<th>Monday</th><th>Tuesday</th><th>Wednesday</th>");
+         printf("<th>Thursday</th><th>Friday</th><th>Saturday</th>");
+         printf("<th>Sunday</th>\n");
+      }
+      else {
+         printf("<th>Sunday</th><th>Monday</th><th>Tuesday</th>");
+         printf("<th>Wednesday</th><th>Thursday</th><th>Friday</th>");
+         printf("<th>Saturday</th>\n");
+
+      }
+      printf("</tr></thead>\n");
+      printf("<tbody><tr>\n");
 
       w = icaltime_day_of_week(day);
-      if (w != 1) {		/* Empty cells until 1st of month */
-	printf("<tr>\n");
-	for (; w > 1; w--) printf("<td class=skip>\n");
+      if (starts_on_monday)
+         if (w == 1) skip = 6; else skip = w-2;
+      else
+         skip = w-1;
+
+      if (skip != 0) {
+         for(; skip > 0; skip--) printf("<td class=\"skip\">&nbsp;</td>\n");
       }
 
       /* Skip events before this day (can only occur at very start) */
@@ -310,16 +400,24 @@ static void print_calendar(const struct icaltimetype start,
 	sprintf(s, "%04d%02d%02d", y, m, d);
 	day = icaltime_from_string(s);
 	w = icaltime_day_of_week(day);
-	if (w == 1) printf("<tr>\n");
-	printf("<td><p class=date>%d\n\n", d);
+	if (w == lastDay && d != 1)
+	{
+		printf("</tr><tr>\n");
+	}
+
+	printf("<td");
+	if (do_today && icaltime_compare_date_only(day, now) == 0)
+	  printf(" id=\"today\"");
+	printf("><p class=\"date\">%d</p>\n\n", d);
 
 	/* Print all events on this day (the events are sorted) */
 	for (; i < nrevents
 	       && icaltime_compare_date_only(events[i].start, day) == 0; i++)
 		print_event(events[i], do_description, do_location);
+	printf("</td>\n");
       }
 
-      printf("</table>\n\n");
+      printf("</tr></tbody></table>\n\n");
     }
 
   }
@@ -332,6 +430,7 @@ static void add_to_queue(icalcomponent *ev, const struct icaltimetype start,
 			 icaltimezone *tz)
 {
   int n = (nrevents/INC + 1) * INC;
+  struct icaltimetype recur_time = icalcomponent_get_recurrenceid(ev);
 
   if (!(events = realloc(events, n * sizeof(*events))))
     fatal(ERR_OUT_OF_MEM, "Out of memory\n");
@@ -339,10 +438,29 @@ static void add_to_queue(icalcomponent *ev, const struct icaltimetype start,
 
   events[nrevents].start = icaltime_convert_to_zone(start, tz);
   events[nrevents].end = icaltime_convert_to_zone(end, tz);
+  events[nrevents].uid = icalcomponent_get_uid(ev);
+  events[nrevents].recur_time = icaltime_convert_to_zone(recur_time, tz);
   events[nrevents].event = ev;
   nrevents++;
 }
 
+/* filter_queue_recurring -- remove obsolete recurring events from queue */
+static void filter_queue_recurring(void)
+{
+  int i, j = 0;
+
+  for (i = 0; i < nrevents; i++) {
+    if (i < nrevents-1 &&
+	strcmp(events[i].uid, events[i+1].uid) == 0 &&
+	icaltime_compare(events[i].start, events[i+1].start) == 0 &&
+	icaltime_is_null_time(events[i].recur_time) &&
+	!icaltime_is_null_time(events[i+1].recur_time))
+      continue;
+    events[j++] = events[i];
+  }
+  nrevents = j;
+}
+
 
 /* iterate -- process each component in the collection */
 static void iterate(icalcomponent *c, struct icaltimetype periodstart,
@@ -444,14 +562,17 @@ int main(int argc, char *argv[])
   icalparser *parser;
   struct icaltimetype periodstart;
   struct icaldurationtype duration;
-  char *footer = NULL, *class = NULL, *category = NULL;
+  char *title = NULL, *footer = NULL, *class = NULL, *category = NULL;
   char *not_class = NULL, *not_category = NULL;
   char c;
   int dummy1, dummy2, dummy3, do_description = 0, do_location = 0;
+  int do_today = 0;
   icaltimezone *tz;
+  struct icaltimetype now;
+  int starts_on_monday = 0;
 
   /* We handle errors ourselves */
-  icalerror_errors_are_fatal = 0;
+  icalerror_set_errors_are_fatal(0);
   icalerror_clear_errno();
 
   /* icaltimezone_set_tzid_prefix("/kde.org/Olson_20080523_1/"); */
@@ -468,8 +589,11 @@ int main(int argc, char *argv[])
     case 'C': not_category = strdup(optarg); break;
     case 'd': do_description = 1; break;
     case 'l': do_location = 1; break;
+    case 't': title = strdup(optarg); break;
     case 'f': footer = strdup(optarg); break;
+    case 'T': do_today = 1; break;
     case 'z': tz = icaltimezone_get_builtin_timezone(optarg); break;
+    case 'm': starts_on_monday = 1; break;
     default: fatal(ERR_USAGE, USAGE);
     }
   }
@@ -477,7 +601,7 @@ int main(int argc, char *argv[])
   if (sscanf(argv[optind], "%04d%02d%02d", &dummy1, &dummy2, &dummy3) < 3)
     fatal(ERR_DATE, "Incorrect date '%s', must be YYYYMMDD.\n", argv[optind]);
   periodstart = icaltime_from_string(argv[optind]);
-  if (icalerrno) ;		/* TO DO */
+  /* if (icalerrno) ;		 TO DO */
   optind++;
   if (optind == argc) fatal(ERR_USAGE, USAGE);
   duration = icaldurationtype_from_string(argv[optind]);
@@ -489,6 +613,7 @@ int main(int argc, char *argv[])
   stream = optind == argc ? stdin : fopen(argv[optind], "r");
   if (!stream) {perror(argv[optind]); exit(1);}
   if (optind + 1 < argc) fatal(ERR_USAGE, USAGE);
+  now = icaltime_current_time_with_zone(tz);
 
   /* Create a new parser object */
   parser = icalparser_new();
@@ -507,9 +632,13 @@ int main(int argc, char *argv[])
   /* Sort the result */
   qsort(events, nrevents, sizeof(*events), compare_events);
 
+  /* Filter out (later) updated recurring event occurrences */
+  filter_queue_recurring();
+
   /* Print the sorted results */
-  print_header(periodstart, duration);
-  print_calendar(periodstart, duration, nrevents, events, do_description, do_location);
+  print_header(periodstart, duration, title);
+  print_calendar(periodstart, duration, nrevents, events, do_description,
+		 do_location, starts_on_monday, do_today, now);
   print_footer(footer);
 
   /* Clean up */
