@@ -183,6 +183,8 @@ static void merge(icalcomponent **a, icalcomponent *b)
   icalproperty *mod_a, *mod_b;
   struct icaltimetype modif_a, modif_b;
   icalproperty *uid;
+  icalproperty *tzid_prop;
+  const char *tzid;
   ENTRY *e, e1;
 
   /* Create the hash table */
@@ -191,6 +193,28 @@ static void merge(icalcomponent **a, icalcomponent *b)
     first_time = 0;
   }
 
+  /* Iterate over VTIMEZONEs */
+  for (h = icalcomponent_get_first_component(b, ICAL_VTIMEZONE_COMPONENT);
+       h; h = next) {
+
+    next = icalcomponent_get_next_component(b, ICAL_VTIMEZONE_COMPONENT);
+
+    tzid_prop = icalcomponent_get_first_property(h, ICAL_TZID_PROPERTY);
+    if (!tzid_prop)
+      continue;
+
+    tzid = icalproperty_get_tzid(tzid_prop);
+    if (!tzid)
+      continue;
+
+    /* Skip existing TZIDs, should use icalcomponent_merge_component() instead. */
+    if (icalcomponent_get_timezone(*a, tzid))
+      continue;
+
+    icalcomponent_remove_component(b, h); /* Move from b... */
+    icalcomponent_add_component(*a, h); /* ... to a */
+  }
+
   /* Iterate over VEVENTs */
   for (h = icalcomponent_get_first_component(b, ICAL_VEVENT_COMPONENT);
        h; h = next) {
@@ -200,6 +224,17 @@ static void merge(icalcomponent **a, icalcomponent *b)
     uid = icalcomponent_get_first_property(h, ICAL_UID_PROPERTY);
     /*debug("%s", uid ? icalproperty_get_uid(uid) : "NO UID!?");*/
     if (!uid) continue;				/* Error in iCalendar file */
+
+    /* Always add VEVENTs with RECURRENCE-ID, as these are events overriding
+     * specific instances of recurring VEVENTs and have the same UID as the
+     * recurring event. We could try to filter VEVENTs with same UID *and*
+     * RECURRENCE-ID. */
+    if (icalcomponent_get_first_property(h, ICAL_RECURRENCEID_PROPERTY)) {
+      icalcomponent_remove_component(b, h); /* Move from b... */
+      icalcomponent_add_component(*a, h); /* ... to a */
+      continue;
+    }
+
     e1.key = strdup(icalproperty_get_uid(uid));
     if (! (e = hsearch(e1, FIND))) {
 
@@ -258,8 +293,8 @@ int main(int argc, char *argv[])
   icalcomponent *newset;
 
   /* We handle errors ourselves */
-  icalerror_errors_are_fatal = 0;
-  icalerrno = 0;
+  icalerror_set_errors_are_fatal(0);
+  icalerror_clear_errno();
 
   /* Read commandline */
   while ((c = getopt_long(argc, argv, OPTIONS, options, NULL)) != -1) {
